(* slide24ctlr.ml     Controller for Slide24 example
 *
 * Copyright (c) 2011 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 *)

(* Slide24ctlr responds to all events, including application state
 * changes.  I.e., Slide24ctlr wears the application delegate hat along
 * with its usual role.
 *)

open Cocoa
open Slide24defs

(* The app can be in five states
 *)
type state =
    | Normal      (* Under user control *)
    | Solving     (* Calculating solution: CPU intensive *)
    | Discouraged (* Under user control, but unable to solve *)
    | Inactive    (* In the background *)
    | Animating   (* Animating solution of puzzle, limited user control *)

(* Options for animation.
 *)
type anim =
    | AnimNone          (* No animation *)
    | AnimNorm          (* Animation with default duration *)
    | AnimDura of float (* Animation with specified duration (sec) *)


class t =
object (self)
    inherit Wrappee.t

    (* These are initialized at launch time.  If they fail to show up,
     * check the NIB file, Slide24.xib.
     *)
    val mutable theWindow: UiWindow.t option = None
    val mutable theShuffleB: UiButton.t option = None
    val mutable theSolveB: UiButton.t option = None

    val mutable tiles: UiButton.t array = [||]
    val mutable config: config = [||]
    val solved = Array.of_list (range 1 24 @ [0])
    val mutable solution: config list = []

    val mutable state: state = Normal

    (* Accessors.
     *)
    method window =
        match theWindow with
        | None -> raise Not_found
        | Some window -> window

    method setWindow' window =
        theWindow <- Some window

    method shuffle =
        match theShuffleB with
        | None -> raise Not_found
        | Some button -> button

    method setShuffle' button =
        theShuffleB <- Some button

    method solve =
        match theSolveB with
        | None -> raise Not_found
        | Some button -> button

    method setSolve' button =
        theSolveB <- Some button

    (* Application state change events.
     *)
    method applicationDidFinishLaunching' (app: UiApplication.t) =
        begin
        (* Startup code.
         *)
        Random.self_init ();
        ignore (
            NsTimer.scheduledTimer5 0.50 self "timerTick'" Wrappee.nil true
        );
        self#make_tiles;
        let styleb bo =
            match bo with
            | None -> ()
            | Some button -> self#style_button button false;
        in
            begin styleb theShuffleB; styleb theSolveB end
        end

    method applicationDidReceiveMemoryWarning' (app: UiApplication.t) =
        Gc.compact ();  (* Best effort to reclaim space *)

    method applicationWillResignActive' (app: UiApplication.t) =
        state <- Inactive

    method applicationDidBecomeActive' (app: UiApplication.t) =
        begin
        state <- Normal;
        self#display;
        end

    method applicationWillTerminate' (app: UiApplication.t) =
        ()

    (* GUI events.
     *)
    method touchTile' (tile: UiButton.t) =
        (* Touched one of the tiles.
         *)
        match state with
        | Normal
        | Discouraged ->
            begin
            match self#rowcol_blocmove (self#rowcol_of_tile tile) with
            | (_, _, []) ->
                () (* Can't move the tile *)
            | (dr, dc, rcs) ->
                let dx, dy = float_of_int dc, float_of_int dr
                in
                    begin
                    UiView.beginAnimations'context' "touchTile" 0;
                    ListLabels.iter rcs ~f:(
                        fun rc ->
                            let t = self#tile_of_rowcol rc
                            in let (x, y, w, h) = t#frame
                            in
                                t#setFrame' (x +. dx *. w, y +. dy *. h , w, h)
                    );
                    UiView.commitAnimations ();
                    ListLabels.iter rcs ~f: (
                        fun (r, c) ->
                            let ix = r * 5 + c
                            in let ix' = ix + dr * 5 + dc
                            in
                                begin
                                config.(ix') <- config.(ix);
                                config.(ix) <- 0;
                                end
                    );
                    state <- Normal;
                    self#display;
                    end
            end
        | _ -> ()

    method shuffle' (b: UiButton.t) =
        match state with
        | Normal
        | Discouraged ->
            begin
            Slide24util.shuffle config 150;
            self#config_tiles (AnimDura 0.6);
            state <- Normal;
            self#display;
            end
        | _ -> ()

    method solve' (b: UiButton.t) =
        begin
        (match state with
        | Normal ->
            begin
            state <- Solving;
            (* Allow the GUI to update before we grab a chunk of CPU time.
             *)
            self#performSelector'withDelay' "runsolve" 0.0;
            end
        | Animating ->
            (* Stop solving.
             *)
            state <- Normal
        | _ -> ()
        );
        self#display;
        end

    method runsolve =
        (* This is rather CPU intensive, so we limit the search.  Every
         * now and then it will fail.
         *)
        begin
        (match Solve.solve config solved 20000 with
        | None ->
            state <- Discouraged
        | Some (soln, _) ->
            begin
            solution <- soln;
            state <- Animating;
            end
        );
        self#display;
        end

    method timerTick' (timer: NsTimer.t) =
        if state = Animating then
            begin
            match solution with
            | [] ->
                begin
                state <- Normal;
                self#display;
                end
            | config' :: rest ->
                begin
                config <- config';
                self#config_tiles AnimNorm;
                solution <- rest;
                end
            end

    (* Private methods.
     *)
    method private make_tiles =
        (* Create tiles at launch.
         *)
        begin
        tiles <- ArrayLabels.init 24 ~f: (
            fun i ->
                let row, col = (float_of_int (i / 5), float_of_int (i mod 5))
                in let tilenum = string_of_int (i + 1)
                in let b = UiButton.buttonWithType' UiButton.TypeCustom
                in let () =
                    begin
                    b#addTarget'action'forControlEvents'
                        self "touchTile'" [UiButton.EventTouchUpInside];
                    b#setFrame'
                        (10.0 +. 60.0 *. col, 30.0 +. 60.0 *. row, 60.0, 60.0);
                    b#setTitleColor'forState'
                        (if i mod 2 = 0 then red else white)
                        UiButton.kStateNormal;
                    b#setTitle'forState' tilenum UiButton.kStateNormal;
                    b#setBackgroundImage'forState'
                        ("tile" ^ tilenum ^ ".png")
                        UiButton.kStateNormal;
                    self#style_button b (i mod 2 = 0);
                    b#setTitleSize' 30.0;
                    match theWindow with
                    | None -> ()
                    | Some w -> w#addSubview' b
                    end
                in
                    b
        );
        config <- Array.copy solved;
        end

    method private style_button b is_white =
        begin
        b#layer#setCornerRadius' 7.0;
        b#layer#setMasksToBounds' true;
        b#layer#setBorderWidth' 2.0;
        b#layer#setBorderColor' (if is_white then dkwhite else dkred);
        end

    method private config_tiles anim =
        (* Move the tiles so they are in the positions given by the
         * current config.
         *)
        begin
        (match anim with
        | AnimNorm
        | AnimDura _ -> UiView.beginAnimations'context' "config_tiles" 0
        | _ -> ()
        );
        (match anim with
        | AnimDura d -> UiView.setAnimationDuration' d
        | _ -> ()
        );
        ArrayLabels.iteri config ~f: (
            fun i tilenum ->
                if tilenum <> 0 then
                    let row, col =
                        (float_of_int (i / 5), float_of_int (i mod 5))
                    in let rect =
                        (10.0 +. 60.0 *. col, 30.0 +. 60.0 *. row, 60.0, 60.0)
                    in let tile = tiles.(tilenum - 1)
                    in
                        tile#setFrame' rect
        );
        (match anim with
        | AnimNorm
        | AnimDura _ -> UiView.commitAnimations ()
        | _ -> ()
        );
        end

    method private rowcol_of_number k : int * int =
        (* Find the row and column of the tile with the given number.
         * The number can also be 0, which returns the row and column of
         * the empty spot.
         *)
        let ix = array_findi ((=) k) config
        in
            (ix / 5, ix mod 5)

    method private rowcol_of_tile (tile: UiButton.t) : int * int =
        (* Return the row and column of the given tile.
         *)
        self#rowcol_of_number (int_of_string tile#currentTitle)

    method private tile_of_rowcol (row, col) : UiButton.t =
        tiles.(config.(row * 5 + col) - 1)

    method private rowcol_blocmove (tr, tc) : int * int * (int * int) list =
        (* If the tile at (tr, tc) can move (possibly en bloc with
         * others), return the unit (row, col) movement vector and the
         * (row, col) values of the tiles that move.  The first entry in
         * the list is the one closest to the empty spot.  If movement
         * is not possible, return a null vector and empty list.
         *)
        let null = (0, 0, [])
        in let vbloc drow col a b =
            (drow, 0, List.map (fun r -> (r, col)) (udrange a b))
        in let hbloc dcol row a b =
            (0, dcol, List.map (fun c -> (row, c)) (udrange a b))
        in let er, ec = self#rowcol_of_number 0
        in
            if tr < er then
                if tc = ec then vbloc 1 tc (er - 1) tr else null
            else if tr > er then
                if tc = ec then vbloc (-1) tc (er + 1) tr else null
            else
                if tc < ec then
                    hbloc 1 tr (ec - 1) tc
                else
                    hbloc (-1) tr (ec + 1) tc

    method private display =
        (* Set the state of the buttons based on the state.
         *)
        match state with
        | Normal ->
            begin
            self#title_buttons "Shuffle" "Solve";
            self#enable_buttons true true;
            end
        | Solving ->
            begin
            self#title_buttons "Shuffle" "Working...";
            self#enable_buttons false false;
            end
        | Discouraged ->
            begin
            self#title_buttons "Shuffle" "Couldn't Solve";
            self#enable_buttons true false;
            end
        | Animating ->
            self#title_buttons "Shuffle" "Stop Solving";
            self#enable_buttons false true;
        | _ -> ()

    method private title_buttons a b =
        let ltitle bo s =
            match bo with
            | None -> ()
            | Some button -> button#setTitle'forState' s UiButton.kStateNormal
        in
            begin ltitle theShuffleB a; ltitle theSolveB b end

    method private enable_buttons a b =
        let lenable bo yesno =
            match bo with
            | None -> ()
            | Some button -> button#setEnabled' yesno
        in
            begin lenable theShuffleB a; lenable theSolveB b end
end

let () =
    let wrapped robjcv =
        let c = new t in let () = c#setContainer robjcv in c
    in
        Callback.register "Slide24ctlr.wrapped" wrapped
